@@ -3,9 +3,10 @@
 //  Software License, Version 1.0. (See accompanying file
 //  LICENSE_1_0.txt or copy at https://www.boost.org/LICENSE_1_0.txt
 
-#ifndef BOOST_MATH_LOGGED_ADAPTER_HPP
-#define BOOST_MATH_LOGGED_ADAPTER_HPP
+#ifndef BOOST_MP_LOGGED_ADAPTER_HPP
+#define BOOST_MP_LOGGED_ADAPTER_HPP
 
+#include <boost/multiprecision/detail/standalone_config.hpp>
 #include <boost/multiprecision/traits/extract_exponent_type.hpp>
 #include <boost/multiprecision/detail/integer_ops.hpp>
 
@@ -83,19 +84,37 @@ struct logged_adaptor
       return *this;
    }
    template <class T>
-   logged_adaptor(const T& i, const typename std::enable_if<std::is_convertible<T, Backend>::value>::type* = 0)
+   logged_adaptor(const T& i, const typename std::enable_if<std::is_convertible<T, Backend>::value>::type* = nullptr)
        : m_value(i)
    {
       log_postfix_event(m_value, "construct from arithmetic type");
    }
    template <class T>
-   logged_adaptor(const logged_adaptor<T>& i, const typename std::enable_if<std::is_convertible<T, Backend>::value>::type* = 0)
+   logged_adaptor(const logged_adaptor<T>& i, const typename std::enable_if<std::is_convertible<T, Backend>::value>::type* = nullptr)
        : m_value(i.value())
    {
       log_postfix_event(m_value, "construct from arithmetic type");
    }
+   template <class T, class U>
+   logged_adaptor(const T& i, const U& j, typename std::enable_if<std::is_constructible<Backend, const T&, const U&>::value>::type* = nullptr)
+      : m_value(i, j)
+   {
+      log_postfix_event(m_value, "construct from a pair of arithmetic types");
+   }
+   template <class D = Backend>
+   logged_adaptor(const Backend& i, unsigned digits10, typename std::enable_if<std::is_constructible<D, Backend const&, unsigned>::value>::type const* = nullptr)
+      : m_value(i, digits10)
+   {
+      log_postfix_event(m_value, "construct from arithmetic type and precision");
+   }
+   template <class D = Backend>
+   logged_adaptor(const logged_adaptor<Backend>& i, unsigned digits10, typename std::enable_if<std::is_constructible<D, Backend const&, unsigned>::value>::type const* = nullptr)
+      : m_value(i.value(), digits10)
+   {
+      log_postfix_event(m_value, "construct from arithmetic type and precision");
+   }
    template <class T>
-   typename std::enable_if<boost::multiprecision::detail::is_arithmetic<T>::value || std::is_convertible<T, Backend>::value, logged_adaptor&>::type operator=(const T& i)
+   typename std::enable_if<boost::multiprecision::detail::is_arithmetic<T>::value || std::is_assignable<Backend, T>::value, logged_adaptor&>::type operator=(const T& i)
    {
       log_prefix_event(m_value, i, "Assignment from arithmetic type");
       m_value = i;
@@ -151,6 +170,8 @@ struct logged_adaptor
    {
       return m_value;
    }
+   
+   #ifndef BOOST_MP_STANDALONE
    template <class Archive>
    void serialize(Archive& ar, const unsigned int /*version*/)
    {
@@ -158,6 +179,8 @@ struct logged_adaptor
       ar& boost::make_nvp("value", m_value);
       log_postfix_event(m_value, "serialize");
    }
+   #endif
+
    static unsigned default_precision() noexcept
    {
       return Backend::default_precision();
@@ -166,6 +189,14 @@ struct logged_adaptor
    {
       Backend::default_precision(v);
    }
+   static unsigned thread_default_precision() noexcept
+   {
+      return Backend::thread_default_precision();
+   }
+   static void thread_default_precision(unsigned v) noexcept
+   {
+      Backend::thread_default_precision(v);
+   }
    unsigned precision() const noexcept
    {
       return value().precision();
@@ -174,6 +205,25 @@ struct logged_adaptor
    {
       value().precision(digits10);
    }
+   //
+   // Variable precision options:
+   // 
+   static constexpr variable_precision_options default_variable_precision_options()noexcept
+   {
+      return Backend::default_variable_precision_options();
+   }
+   static constexpr variable_precision_options thread_default_variable_precision_options()noexcept
+   {
+      return Backend::thread_default_variable_precision_options();
+   }
+   static BOOST_MP_CXX14_CONSTEXPR void default_variable_precision_options(variable_precision_options opts)
+   {
+      Backend::default_variable_precision_options(opts);
+   }
+   static BOOST_MP_CXX14_CONSTEXPR void thread_default_variable_precision_options(variable_precision_options opts)
+   {
+      Backend::thread_default_variable_precision_options(opts);
+   }
 };
 
 template <class T>
@@ -320,6 +370,49 @@ inline void eval_convert_to(R* result, const logged_adaptor<Backend>& val)
    log_postfix_event(val.value(), *result, "convert_to");
 }
 
+template <class Backend, class R>
+inline void eval_convert_to(logged_adaptor<R>* result, const logged_adaptor<Backend>& val)
+{
+   using default_ops::eval_convert_to;
+   log_prefix_event(val.value(), "convert_to");
+   eval_convert_to(&result->value(), val.value());
+   log_postfix_event(val.value(), &result->value(), "convert_to");
+}
+template <class Backend, class R>
+inline void eval_convert_to(logged_adaptor<R>* result, const Backend& val)
+{
+   using default_ops::eval_convert_to;
+   log_prefix_event(val, "convert_to");
+   eval_convert_to(&result->value(), val);
+   log_postfix_event(val, &result->value(), "convert_to");
+}
+
+template <class Backend>
+inline void eval_convert_to(std::complex<float>* result, const logged_adaptor<Backend>& val)
+{
+   using default_ops::eval_convert_to;
+   log_prefix_event(val.value(), "convert_to");
+   eval_convert_to(result, val.value());
+   log_postfix_event(val.value(), *result, "convert_to");
+}
+template <class Backend>
+inline void eval_convert_to(std::complex<double>* result, const logged_adaptor<Backend>& val)
+{
+   using default_ops::eval_convert_to;
+   log_prefix_event(val.value(), "convert_to");
+   eval_convert_to(result, val.value());
+   log_postfix_event(val.value(), *result, "convert_to");
+}
+template <class Backend>
+inline void eval_convert_to(std::complex<long double>* result, const logged_adaptor<Backend>& val)
+{
+   using default_ops::eval_convert_to;
+   log_prefix_event(val.value(), "convert_to");
+   eval_convert_to(result, val.value());
+   log_postfix_event(val.value(), *result, "convert_to");
+}
+
+
 template <class Backend, class Exp>
 inline void eval_frexp(logged_adaptor<Backend>& result, const logged_adaptor<Backend>& arg, Exp* exp)
 {
@@ -339,6 +432,7 @@ inline void eval_ldexp(logged_adaptor<Backend>& result, const logged_adaptor<Bac
 template <class Backend, class Exp>
 inline void eval_scalbn(logged_adaptor<Backend>& result, const logged_adaptor<Backend>& arg, Exp exp)
 {
+   using default_ops::eval_scalbn;
    log_prefix_event(arg.value(), "scalbn");
    eval_scalbn(result.value(), arg.value(), exp);
    log_postfix_event(result.value(), exp, "scalbn");
@@ -347,6 +441,7 @@ inline void eval_scalbn(logged_adaptor<Backend>& result, const logged_adaptor<Ba
 template <class Backend>
 inline typename Backend::exponent_type eval_ilogb(const logged_adaptor<Backend>& arg)
 {
+   using default_ops::eval_ilogb;
    log_prefix_event(arg.value(), "ilogb");
    typename Backend::exponent_type r = eval_ilogb(arg.value());
    log_postfix_event(arg.value(), "ilogb");
@@ -436,37 +531,37 @@ inline void eval_right_shift(logged_adaptor<Backend>& arg, const logged_adaptor<
 }
 
 template <class Backend, class T>
-inline unsigned eval_integer_modulus(const logged_adaptor<Backend>& arg, const T& a)
+inline T eval_integer_modulus(const logged_adaptor<Backend>& arg, const T& a)
 {
    using default_ops::eval_integer_modulus;
    log_prefix_event(arg.value(), a, "integer-modulus");
-   unsigned r = eval_integer_modulus(arg.value(), a);
+   T r = eval_integer_modulus(arg.value(), a);
    log_postfix_event(arg.value(), r, "integer-modulus");
    return r;
 }
 
 template <class Backend>
-inline unsigned eval_lsb(const logged_adaptor<Backend>& arg)
+inline std::size_t eval_lsb(const logged_adaptor<Backend>& arg)
 {
    using default_ops::eval_lsb;
    log_prefix_event(arg.value(), "least-significant-bit");
-   unsigned r = eval_lsb(arg.value());
+   std::size_t r = eval_lsb(arg.value());
    log_postfix_event(arg.value(), r, "least-significant-bit");
    return r;
 }
 
 template <class Backend>
-inline unsigned eval_msb(const logged_adaptor<Backend>& arg)
+inline std::size_t eval_msb(const logged_adaptor<Backend>& arg)
 {
    using default_ops::eval_msb;
    log_prefix_event(arg.value(), "most-significant-bit");
-   unsigned r = eval_msb(arg.value());
+   std::size_t r = eval_msb(arg.value());
    log_postfix_event(arg.value(), r, "most-significant-bit");
    return r;
 }
 
 template <class Backend>
-inline bool eval_bit_test(const logged_adaptor<Backend>& arg, unsigned a)
+inline bool eval_bit_test(const logged_adaptor<Backend>& arg, std::size_t a)
 {
    using default_ops::eval_bit_test;
    log_prefix_event(arg.value(), a, "bit-test");
@@ -476,7 +571,7 @@ inline bool eval_bit_test(const logged_adaptor<Backend>& arg, unsigned a)
 }
 
 template <class Backend>
-inline void eval_bit_set(const logged_adaptor<Backend>& arg, unsigned a)
+inline void eval_bit_set(const logged_adaptor<Backend>& arg, std::size_t a)
 {
    using default_ops::eval_bit_set;
    log_prefix_event(arg.value(), a, "bit-set");
@@ -484,7 +579,7 @@ inline void eval_bit_set(const logged_adaptor<Backend>& arg, unsigned a)
    log_postfix_event(arg.value(), arg, "bit-set");
 }
 template <class Backend>
-inline void eval_bit_unset(const logged_adaptor<Backend>& arg, unsigned a)
+inline void eval_bit_unset(const logged_adaptor<Backend>& arg, std::size_t a)
 {
    using default_ops::eval_bit_unset;
    log_prefix_event(arg.value(), a, "bit-unset");
@@ -492,7 +587,7 @@ inline void eval_bit_unset(const logged_adaptor<Backend>& arg, unsigned a)
    log_postfix_event(arg.value(), arg, "bit-unset");
 }
 template <class Backend>
-inline void eval_bit_flip(const logged_adaptor<Backend>& arg, unsigned a)
+inline void eval_bit_flip(const logged_adaptor<Backend>& arg, std::size_t a)
 {
    using default_ops::eval_bit_flip;
    log_prefix_event(arg.value(), a, "bit-flip");
@@ -537,6 +632,10 @@ NON_MEMBER_OP2(logb, "logb")
 NON_MEMBER_OP3(fmod, "fmod")
 NON_MEMBER_OP3(pow, "pow")
 NON_MEMBER_OP3(atan2, "atan2")
+NON_MEMBER_OP2(asinh, "asinh")
+NON_MEMBER_OP2(acosh, "acosh")
+NON_MEMBER_OP2(atanh, "atanh")
+NON_MEMBER_OP2(conj, "conj")
 
 template <class Backend>
 int eval_signbit(const logged_adaptor<Backend>& val)
@@ -551,6 +650,72 @@ std::size_t hash_value(const logged_adaptor<Backend>& val)
    return hash_value(val.value());
 }
 
+template <class Backend, expression_template_option ExpressionTemplates>
+inline typename std::enable_if<number_category<Backend>::value == number_kind_rational, typename number<logged_adaptor<Backend>, ExpressionTemplates>::value_type>::type
+numerator(const number<logged_adaptor<Backend>, ExpressionTemplates>& arg)
+{
+   number<Backend, ExpressionTemplates> t(arg.backend().value());
+   return numerator(t).backend();
+}
+template <class Backend, expression_template_option ExpressionTemplates>
+inline typename std::enable_if<number_category<Backend>::value == number_kind_rational, typename number<logged_adaptor<Backend>, ExpressionTemplates>::value_type>::type
+denominator(const number<logged_adaptor<Backend>, ExpressionTemplates>& arg)
+{
+   number<Backend, ExpressionTemplates> t(arg.backend().value());
+   return denominator(t).backend();
+}
+
+template <class To, class From>
+inline BOOST_MP_CXX14_CONSTEXPR void eval_set_real(To& to, const logged_adaptor<From>& from)
+{
+   using default_ops::eval_set_real;
+   log_prefix_event(to, from.value(), "Set real part");
+   eval_set_real(to, from.value());
+   log_postfix_event(to, from.value(), "Set real part");
+}
+template <class To, class From>
+inline BOOST_MP_CXX14_CONSTEXPR void eval_set_real(logged_adaptor<To>& to, const logged_adaptor<From>& from)
+{
+   using default_ops::eval_set_real;
+   log_prefix_event(to.value(), from.value(), "Set real part");
+   eval_set_real(to.value(), from.value());
+   log_postfix_event(to.value(), from.value(), "Set real part");
+}
+template <class To, class From>
+inline BOOST_MP_CXX14_CONSTEXPR void eval_set_real(logged_adaptor<To>& to, const From& from)
+{
+   using default_ops::eval_set_real;
+   log_prefix_event(to.value(), from, "Set real part");
+   eval_set_real(to.value(), from);
+   log_postfix_event(to.value(), from, "Set real part");
+}
+
+template <class To, class From>
+inline BOOST_MP_CXX14_CONSTEXPR void eval_set_imag(To& to, const logged_adaptor<From>& from)
+{
+   using default_ops::eval_set_imag;
+   log_prefix_event(to, from.value(), "Set imag part");
+   eval_set_imag(to, from.value());
+   log_postfix_event(to, from.value(), "Set imag part");
+}
+template <class To, class From>
+inline BOOST_MP_CXX14_CONSTEXPR void eval_set_imag(logged_adaptor<To>& to, const logged_adaptor<From>& from)
+{
+   using default_ops::eval_set_imag;
+   log_prefix_event(to.value(), from.value(), "Set imag part");
+   eval_set_imag(to.value(), from.value());
+   log_postfix_event(to.value(), from.value(), "Set imag part");
+}
+template <class To, class From>
+inline BOOST_MP_CXX14_CONSTEXPR void eval_set_imag(logged_adaptor<To>& to, const From& from)
+{
+   using default_ops::eval_set_imag;
+   log_prefix_event(to.value(), from, "Set imag part");
+   eval_set_imag(to.value(), from);
+   log_postfix_event(to.value(), from, "Set imag part");
+}
+
+
 #define NON_MEMBER_COMPLEX_TO_REAL(name, str)                                                    \
    template <class B1, class B2>                                                                 \
    inline void BOOST_JOIN(eval_, name)(logged_adaptor<B1> & result, const logged_adaptor<B2>& a) \
@@ -577,17 +742,54 @@ NON_MEMBER_COMPLEX_TO_REAL(imag, "imag")
 template <class T, class V, class U>
 inline void assign_components(logged_adaptor<T>& result, const V& v1, const U& v2)
 {
-   assign_components(result.value(), v1, v2);
+   using default_ops::assign_components;
+   assign_components(result.value(), unwrap_logged_type(v1), unwrap_logged_type(v2));
 }
 
 } // namespace backends
 
 using backends::logged_adaptor;
 
+namespace detail {
+   template <class Backend>
+   struct is_variable_precision<logged_adaptor<Backend> > : public is_variable_precision<Backend>
+   {};
+#ifdef BOOST_HAS_INT128
+   template <class Backend>
+   struct is_convertible_arithmetic<int128_type, logged_adaptor<Backend> > : public is_convertible_arithmetic<int128_type, Backend>
+   {};
+   template <class Backend>
+   struct is_convertible_arithmetic<uint128_type, logged_adaptor<Backend> > : public is_convertible_arithmetic<uint128_type, Backend>
+   {};
+#endif
+#ifdef BOOST_HAS_FLOAT128
+   template <class Backend>
+   struct is_convertible_arithmetic<float128_type, logged_adaptor<Backend> > : public is_convertible_arithmetic<float128_type, Backend>
+   {};
+#endif
+   } // namespace detail
+
 template <class Backend>
 struct number_category<backends::logged_adaptor<Backend> > : public number_category<Backend>
 {};
 
+template <class Number>
+using logged_adaptor_t = number<logged_adaptor<typename Number::backend_type>, Number::et>;
+
+template <class Backend, expression_template_option ExpressionTemplates>
+struct component_type<number<logged_adaptor<Backend>, ExpressionTemplates>>
+{
+   //
+   // We'll make the component_type also a logged_adaptor:
+   //
+   using base_component_type = typename component_type<number<Backend, ExpressionTemplates>>::type;
+   using base_component_backend = typename base_component_type::backend_type;
+   using type = number<logged_adaptor<base_component_backend>, ExpressionTemplates>;
+};
+
+template <class Backend>
+struct is_interval_number<backends::logged_adaptor<Backend> > : public is_interval_number<Backend> {};
+
 }} // namespace boost::multiprecision
 
 namespace std {
@@ -613,6 +815,7 @@ class numeric_limits<boost::multiprecision::number<boost::multiprecision::backen
 
 } // namespace std
 
+#ifdef BOOST_MP_MATH_AVAILABLE
 namespace boost {
 namespace math {
 
@@ -626,6 +829,7 @@ struct precision<boost::multiprecision::number<boost::multiprecision::logged_ada
 }
 
 }} // namespace boost::math::policies
+#endif // BOOST_MP_MATH_AVAILABLE
 
 #undef NON_MEMBER_OP1
 #undef NON_MEMBER_OP2
